/*
 * The Helium Toolkit
 *
 * Copyright (C) 1996-2000 Edouard Thiel <Edouard.Thiel@lif.univ-mrs.fr>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License; see http://www.lif-sud.univ-mrs.fr/~thiel/helium/LICENSE
*/

/*
 * gui/destroy.c - 04/07/1999
 *
 * 
*/

#include <helium.h>

int he_exit_code = 0;

static int he_destroying_flag = FALSE; 
static He_vlist he_destroy_list;


/*------------------- I N T E R F A C E - P U B L I Q U E --------------------*/


/*
 * Enregistre hn dans he_destroy_list pour destroy differe
*/

void HeDestroy (He_node *hn)
{
    int redraw_sub;
    
    if (he_destroying_flag) {
	HeError ("HeDestroy: call forbidden in destroy_proc\n");
	return;
    } 
    
    if (hn == NULL || HE_BIT_ISSET (hn->status, HE_STA_DESTROY)) return;
    
    if (HeVListInsert (&he_destroy_list, &hn) < 0) return;
    
    HE_BIT_SET (hn->status, HE_STA_DESTROY);
    redraw_sub = !HE_BIT_ISSET (hn->meth->props, HE_PM_WIN_OWNER);
    
    /* On marque tout de suite les fils :
     *   - pour avoir une info immediate avant le HeDestroyFlush 
     *    (utilisee dans HeBoundingBox) ;
     *   - pour eviter par la suite de detruire 2 fois le meme noeud.
    */
    HeDestroyPropagateFlag (hn, redraw_sub);
}


/*
 * Detruit tous les objets depuis la racine. Fait sortir proprement 
 * de HeIOLoop() et HeMainLoop() en renvoyant exit_code.
 * (le display est ferme' dans HeMainLoop() juste avant la sortie).
 *
 * Si HeQuit est appele avant HeMainLoop, referme le display et
 * fait exit(exit_code).
 * Si HeQuit est appele apres HeMainLoop, fait exit(exit_code).
*/

void HeQuit (int exit_code)
{
    /* Si c'est un appel depuis une destroy_proc */
    if (he_destroying_flag) return;
    
    /* Mmorise exit_code dans variable globale */
    he_exit_code = exit_code;

    if (he_root_node) HeDestroy (he_root_node);

    if (he_main_frame) {
	/* Fait sortir de HeMainLoop */
	HeIOLoopStop ();
    } else { 
        /* On n'est pas dans HeMainLoop */
	if (he_display) XCloseDisplay (he_display);
        exit(exit_code);
    }
}


/*-------------------- F O N C T I O N S - P R I V E E S ---------------------*/

/*
 * Initialisation generale du module.
*/

void HeDestroyInit ()
{
   /* Liste de He_node* non trie
    * stockage dans un ordre quelconque.
    * suppression en bouchant le trou avec le dernier lment.
   */
    HeVListInit (&he_destroy_list, FALSE, 128, sizeof(He_node*), NULL);
}


/*
 * Ralisation des Destroy ; appel des destroy_proc.
 *
*/

void HeDestroyFlush ()
{
    int i, res = 0;

    if (he_destroy_list.nb == 0) return;
    
#   ifdef DEBUG
    printf ("HeDestroyFlush : nb = %d  max = %d\n", 
        he_destroy_list.nb, he_destroy_list.max);
#   endif

    /* Suppression de la liste des noeuds deja dans un sous-arbre 
     * pour eviter de detruire 2 fois le meme noeud. 
     * En cas de suppression, on fait i-- pour repasser sur l'elt qui 
     * bouche le trou dans la vlist.
    */
    for (i = 0; i < he_destroy_list.nb; i++) {
        He_node *hn = ((He_node**)he_destroy_list.v)[i];
 	if (HE_BIT_ISSET (hn->status, HE_STA_DESTROY_INHERIT))
            HeVListRemove (&he_destroy_list, i--);
    }
                
    /* Appel des destroy_proc : on interdit l'appel de HeDestroy() */
    he_destroying_flag = TRUE;
    for (i = 0; i < he_destroy_list.nb; i++) {
        He_node *hn = ((He_node**)he_destroy_list.v)[i];
        HeDestroyCallProc (hn);
    }
    he_destroying_flag = FALSE;
    
    /* Masquage et destruction de bas en haut */
    for (i = 0; i < he_destroy_list.nb; i++) {
        He_node *hn = ((He_node**)he_destroy_list.v)[i];
        if (HeDestroyRec (hn) < 0) res = -1;
    }
    
    if (res < 0) 
	HeError ("HeDestroyFlush: error during delayed destroy");
    
    /* RAZ liste */
    HeVListRAZ (&he_destroy_list);       
}


/*
 * Propage destroy, destroy_inherit et redraw_sub dans tous les fils
*/

void HeDestroyPropagateFlag (He_node *hn, int redraw_sub)
{
    if (hn == NULL) return;
    
    /* Parcours des fils */
    for (hn = hn->sub_node; hn != NULL ; hn = hn->next_node) {
    
        /* On marque le destroy differe + herite' d'un pere */
	HE_BIT_SET (hn->status, HE_STA_DESTROY);
        HE_BIT_SET (hn->status, HE_STA_DESTROY_INHERIT);
        
        /* On enleve show pour ne pas redessiner le fond */
        if (!redraw_sub) HE_BIT_CLR (hn->status, HE_STA_SHOW);
        
        /* On propage FALSE si hn est un window, redraw_sub sinon */
        HeDestroyPropagateFlag (hn,
            redraw_sub && !HE_BIT_ISSET (hn->meth->props, HE_PM_WIN_OWNER));
    }
}


/*
 * Appelle la destroy_proc de hn puis de ses fils
*/

void HeDestroyCallProc (He_node *hn)
{
    if (hn == NULL) return;
    
    if (hn->destroy_proc != NULL)
	(hn->destroy_proc)(hn);
        
    for (hn = hn->sub_node; hn != NULL ; hn = hn->next_node)
        HeDestroyCallProc (hn);
}


/*
 * Marquage et destruction de hn et de ses fils
 *
 * Renvoie 0 succes, ou -1 erreur, mais sans bloquer :
 * le but est de detruire tout ce qu'on peut.
*/

int HeDestroyRec (He_node *hn)
{
    int res = 0;
    He_node *hop, *next;
    
    if (hn == NULL) return 0;
    
    /* Masquage
    */
    if (HE_BIT_ISSET (hn->status, HE_STA_SHOW)) {
        if (HE_BIT_ISSET (hn->meth->props, HE_PM_WIN_OWNER))
            XUnmapWindow (he_display, hn->win);
        else {
            HE_BIT_CLR (hn->status, HE_STA_SHOW);
            hn->meth->draw (hn);
        }
    }

    /* Appel de HeDestroyRec sur les sub_node.
     * Dans cet ordre, on detruit en remontant.
    */
    hop = hn->sub_node;
    while (hop != NULL) {
	/* Attention il faut sauver hop->next_node avant de detruire hop */
	next = hop->next_node;
	if ( HeDestroyRec (hop) < 0 ) res = -1;
        hop = next;
    }

    /* Appel de la methode He..Destroy (sans bloquer sur une erreur).
     * On y detruit tous les champs de l'objet, eventuellement son Window
     * puis on detruit l'objet.
    */
    if (hn != he_root_node)
	(hn->meth->destroy)(hn);

    /* Detache hn de l'arbre des noeuds.
     * Operation possible que si le fils est NULL et le pere non NULL.
    */
    if (hn != he_root_node)
	if ( HeNodeRemoveTree (hn) < 0 ) res = -1;

    /* Destruction du noeud */
    if ( HeNodeFree (hn) < 0 ) res = -1;

    return res;
}


