/*
 * The Helium Toolkit
 *
 * Copyright (C) 1996-2000 Edouard Thiel <Edouard.Thiel@lif.univ-mrs.fr>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License; see http://www.lif-sud.univ-mrs.fr/~thiel/helium/LICENSE
*/

/*
 * gui/noeud.c - 28/03/2000
 *
 * 
*/

#include <helium.h>

/* Liste de noeuds trie dans l'ordre croissant des id */
static He_vlist he_id_list;

static int he_node_uniq;  	/* numero pour id unique */
           
He_node *he_root_node = NULL;


/*------------------- I N T E R F A C E - P U B L I Q U E --------------------*/


/*
 * Recherche rapide d'un He_node  partir de son id.
 * Renvoie *noeud ou NULL.
 * Silencieux.
*/

He_node *HeGetNodeFromId (int id)
{
    He_node st_elt, *elt = &st_elt;
    int index, res;
    
    elt->id = id;
    res = HeVListSearch (&he_id_list, &elt, &index);
    return (res == 0) ? ((He_node**)he_id_list.v)[index] : NULL;
}


/*
 * Renvoie le noeud a l'indice i dans node_list, ou NULL
 * Utilite : parcours de tous les noeuds.
*/

He_node *HeNodeScan (int i)
{
    return (i < 0 || i >= he_id_list.nb) ? NULL : ((He_node**)he_id_list.v)[i];
}


/*-------------------- F O N C T I O N S - P R I V E E S ---------------------*/

/*
 * Initialisation generale du module.
 * Retour erreur -1, succes 0.
*/

int HeNodeInit ()
{
    HeVListInit (&he_id_list, TRUE, 256, sizeof(He_node*), HeNodeComparId);

    he_node_uniq = 10000;

    /* Cre'ation du noeud "top" he_root_node */
    he_root_node = HeNodeNew();
    if (he_root_node == NULL) return -1;
    HE_BIT_CLR (he_root_node->status, HE_STA_SHOW);
    he_root_node->xb = 0;
    he_root_node->yb = 0;
    he_root_node->xm = he_root_width;
    he_root_node->ym = he_root_height;
    
    return 0;
}


/*
 * Callback de comparaison pour HeVListSearch()
*/

int HeNodeComparId (void *v1, void *v2)
{
    return (*(He_node **)v1)->id == (*(He_node **)v2)->id ?  0 :
           (*(He_node **)v1)->id <  (*(He_node **)v2)->id ? -1 : 1;
}


/*
 * Creation d'un nouveau noeud avec id unique
 * Insertion dans he_id_list, agrandissement automatique
 * Renvoie *noeud ou NULL.
*/

He_node *HeNodeNew ()
{
    /* Allocation memoire */
    He_node *hn = malloc (sizeof (He_node));

    /* Test succes */
    if (hn == NULL) {
	HeError ("HeNodeNew: malloc error\n");
	return (He_node *) NULL;
    }

    /* Init les champs, dont le champ hn->id */
    HeBaseInit (hn, he_node_uniq++);

    /* Attachement dans he_id_list (agrandissement automatique) */
    if (HeVListInsert (&he_id_list, &hn) < 0) {
        HeBaseZero (hn);
        free (hn);
        return NULL;
    }

    /* Fin */
    return hn;
}


/*
 * Cherche la position du noeud dans he_id_list, libere le noeud,
 * m-a-j de he_id_list, rduction automatique
 * Retour erreur -1, succes 0.
 *
 * ATTENTION il faut avoir libere' les sous-champs du noeud AVANT. 
*/

int HeNodeFree (He_node *hn)
{
    int index;

    if (hn == NULL) return 0;
    
    /* Cherche la position du noeud dans he_id_list  partir de son id.
    */
    if (HeVListSearch (&he_id_list, &hn, &index) < 0) {
	HeError ("HeNodeFree: object %d doesn't exist\n", hn->id);
	return -1;
    }

    /* Suppression de la liste (rduction automatique) */
    if (HeVListRemove (&he_id_list, index) < 0) return -1;
    
    /* Tout est ok, on libere le noeud */
    HeBaseZero (hn);
    free (hn);

    return 0;
}


/*
 * Attache un noeud a un pere
*/

void HeNodeAddTree (He_node *owner, He_node *nouv)
{
    He_node *prem = NULL;

    if (nouv == NULL) {
	HeError ("HeNodeAddTree: node is NULL\n");
	return;
    }

    if (owner != NULL) {
	prem = owner->sub_node;
	owner->sub_node = nouv;
    }
    nouv->owner = owner;
    nouv->sub_node = NULL;
    nouv->next_node = prem;
}

/*
 * Detache un noeud de son pere et de ses freres.
 * Operation possible que si le fils est NULL et le pere non NULL.
 * Renvoie succes 0, erreur -1.
*/

int HeNodeRemoveTree (He_node *hn)
{
    He_node *cur = NULL;

    if (hn == NULL) {
	HeError ("HeNodeRemoveTree: node is NULL\n");
	return -1;
    }
    if (hn->sub_node != NULL) {
	HeError ("HeNodeRemoveTree: sub_node is not NULL\n");
	return -1;
    }
    if (hn->owner == NULL) {
	HeError ("HeNodeRemoveTree: owner is NULL\n");
	return -1;
    }
    cur = hn->owner->sub_node;
    if (cur == NULL) {
	HeError ("HeNodeRemoveTree: sub_node of owner is NULL\n");
	return -1;
    }
    if (cur == hn) {
	hn->owner->sub_node = cur->next_node;
	return 0;
    }
    while (cur->next_node != NULL && cur->next_node != hn)
	cur = cur->next_node;
    if (cur == NULL) {
	HeError ("HeNodeRemoveTree:"
                 " node not founded in sub_node list of its owner\n");
	return -1;
    }
    cur->next_node = hn->next_node;
    return 0;
}


/*
 * Affiche infos pour debugage
*/

void HeNodeDebug ()
{
    int i;

    printf ("node_nb = %d   node_max = %d\n", 
        he_id_list.nb, he_id_list.max);

    for (i = 0; i < he_id_list.nb; i++)
	printf ("  [%d]  id = %d\n", 
            i, ((He_node**)he_id_list.v)[i]->id);
}


