/*
 * The Helium Toolkit
 *
 * Copyright (C) 1996-2000 Edouard Thiel <Edouard.Thiel@lif.univ-mrs.fr>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License; see http://www.lif-sud.univ-mrs.fr/~thiel/helium/LICENSE
*/

/*
 * gui/principal.c - 30/05/1996
 *
 * 
*/

#include <helium.h>


/* 
 * Variables globales et exportees
*/
Display *he_display;
int he_screen;
Window he_root_window;
int he_root_width, he_root_height;
int he_default_depth;
Visual *he_default_visual;
Colormap he_default_cmap;
int he_init_stat = 0;
He_node *he_main_frame = NULL;
Time he_multiclic_timeout = 400;  /* millisecondes */
char *he_base_name = NULL, *he_res_name = NULL, *he_res_class = NULL,
     *he_debug_option = NULL;


/*------------------- I N T E R F A C E - P U B L I Q U E --------------------*/

/*
 * Initialisation generale de HELIUM et ouverture du display X.
 * On lui transmet les arguments de la ligne de commande ; toutes les 
 * options reconnues sont substituees de argv.
 * 
 * Appelle HeQuit(exit_code) en cas d'erreur.
*/

void HeInit (int *argc, char ***argv)
{
    char *display_name;
    
    /*
     * Calcul du nom de l'application
    */
    if (he_base_name == NULL) {
        int i, n = 0;
        for (i = 0; (*argv)[0][i]; i++)
            if ((*argv)[0][i] == '/') n = i+1;
        he_base_name = strdup ((*argv)[0] + n);
    }
    he_res_name = he_base_name;
    he_res_class = "Helium";
    
    
    /* 
     * Lecture des arguments de la ligne de commande et des ressources
    */
    HeDeclArg ("-help", NULL, NULL, 0, "print options");
    HeDeclArg ("-res", NULL, NULL, 0, "print X resources");
    HeDeclArg ("-debug", NULL, "<option>", 1, "print debug infos");
    HeDeclArg ("-display", NULL, "<display name:0>", 1, NULL);
    HeDeclArg ("-fn", "fontNormal", "<normal font name>", 1, NULL);
    HeDeclArg ("-fb", "fontBold", "<bold font name>", 1, NULL);
    HeDeclArg ("-fl", "fontSmall", "<small font name>", 1, NULL);
    HeDeclArg ("-fs", "fontSize", "<default fonts size>", 1, NULL);
    HeDeclArg ("-geom", "winGeom", "<WxH+X+Y>", 1, "set window geometry");
    HeDeclArg ("-color", "winColor", "<color name>", 1, "set window color");
    HeDeclArg (NULL, "selectColor", "<color name>", 1, "set selection color");
    HeDeclArg (NULL, "toggleColor", "<color name>", 1, "set toggle button color");
    
    /* On interdit dorenavant d'appeler HeDeclArg() */
    he_init_stat = 1; 
    
    if (HeAnalyseArgs (argc, argv) < 0) HeQuit(1);
    
    /* Autorise appels a HeGetOption() */
    he_init_stat = 2; 

    if (HeGetOption ("-help", NULL)) { HePrintHelp (); HeQuit(0); }
    if (HeGetOption ("-res", NULL)) { HePrintRes (); HeQuit(0); }

    /* Recherche du display_name */
    display_name = HeGetOption ("-display", NULL);
    if (display_name == NULL) display_name = getenv("DISPLAY");
    if (display_name == NULL || display_name[0] == 0) display_name = ":0";

    /*
     * Initialisation de X11
    */
    he_display = XOpenDisplay (display_name);
    if (he_display == NULL) {
	HeError ("HeInit: XOpenDisplay(\"%s\") failed\n", display_name);
	HeQuit(1);
    }
    /* Detourne les erreurs */
    XSetErrorHandler(HeNonFatalXErrorHandler);

    he_screen         = DefaultScreen (he_display);
    he_root_window    = RootWindow    (he_display, he_screen);
    he_root_width     = DisplayWidth  (he_display, he_screen);
    he_root_height    = DisplayHeight (he_display, he_screen);
    he_default_depth  = DefaultDepth  (he_display, he_screen);
    he_default_visual = DefaultVisual (he_display, he_screen);
    he_default_cmap   = DefaultColormap (he_display, he_screen);

    /*
     * Initialisation des couleurs, des fontes et des GC
    */
    HeColorSysInit ();
    HeInitGCs ();
    HeXiInit (FALSE);
    if (HeInitFonts () < 0) HeQuit(1);

    /*
     * Initialisation des modules 
    */
    if (HeNodeInit () < 0) HeQuit(1);
    HeWinListInit ();
    HeUpdateInit ();
    HeDestroyInit ();
    HeAtomInit ();
    HeMethodInit ();
    HeIOInit ();
    HeTimerInit ();
    HeSignalInit ();
    
    /*
     * Enregistrement du fd li au display X11
    */
    HeEventIOInit ();
    
    /* Calcul de he_wmdeco_width, he_wmdeco_height */
    HeFrameReadAtomWMDeco ();
    
    /*
     * Informations de debuggage
    */
    he_debug_option = HeGetOption ("-debug", NULL);
    if (he_debug_option) {
        if (strcmp(he_debug_option, "xi") == 0) {
	    HeXiDebug();
        } else if (strcmp(he_debug_option, "wmdeco") == 0) {
            /* utilis plus tard */
        } else {
            printf ("Debug options : xi wmdeco\n");
        }
    }
}


/*
 * Boucle principale pseudo-infinie d'attente des evenements.
 * On lui transmet le frame principal qui sera affiche'.
 *
 * Renvoie code de sortie provenant de HeQuit(exit_code).
 *
 * Attention, apres HeMainLoop() le display est ferme' et on ne peut plus
 * faire aucune operation X. Cet emplacement est destine' a la fermeture
 * des fichiers.
 *
 * Appelle HeQuit() en cas d'erreur.
*/

int HeMainLoop (He_node *frame)
{
    if (he_init_stat != 2) {
	HeError ("HeMainLoop ignored: multiple HeMainLoop() calls"
                 " are forbidden. To show a frame, use HeSetShow insteed\n");
	return -1;
    }
    /* Interdit nouvel appel  HeMainLoop */
    he_init_stat = 3; 
    
    /* Memorisation et affichage du frame principal */
    if (HeSetMainFrame (frame) < 0) HeQuit(1);
    
    /* Ralise affichages diffrs + XFlush */
    HeFlushDisplay ();
    
    /* Met boite de dialogue eventuelle au premier plan :( */
    HeDialogOnTop ();

    /* Boucle d'evenements infinie.
     * Un appel a HeQuit(exit_code) fait sortir proprement de HeIOLoop()
     * et met le code de sortie dans la variable globale he_exit_code.
    */
    HeIOLoop ();
    
    /* Tous les noeuds ont ete detruits.
     * On ferme maintenant le display X.
     * ATTENTION : ne plus faire aucune operation X apres.
    */
    XCloseDisplay (he_display); 
    he_display = NULL;
    he_main_frame = NULL;
    he_root_node = NULL;
    
    return he_exit_code;
}


/*-------------------- F O N C T I O N S - P R I V E E S ---------------------*/

/*
 * Memorisation et affichage du frame principal.
 *
 * Renvoie -1 erreur, 0 succes.
*/

int HeSetMainFrame (He_node *frame)
{
    int i;
    He_node *hop;
    
    if (frame == NULL) {
	HeError ("HeMainLoop: NULL arg\n");
	return -1;
    }

    /* On teste si c'est bien un FRAME */
    if (!HeIsFrame(frame)) {
	HeError ("HeMainLoop: arg is not a frame\n");
	return -1;
    }
    
    /* Memorisation du frame principal */
    he_main_frame = frame;
    
    /* On donne aux frames deja crees le Leader du Window Group */
    for (i = 0; (hop = HeNodeScan(i)) != NULL; i++)
	if (HeIsFrame(hop))
	    HeWMGroup (hop);
                    
    /* On applique l'option "-geom" avant que la fenetre ne soit mappee */
    HeFrameOptionGeom ();
    
    /* Mappage du frame principal */
    HeSetShow (he_main_frame, TRUE);
    
    return 0;
}


