/*
 * The Helium Toolkit
 *
 * Copyright (C) 1996-2000 Edouard Thiel <Edouard.Thiel@lif.univ-mrs.fr>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License; see http://www.lif-sud.univ-mrs.fr/~thiel/helium/LICENSE
*/

/*
 * gui/timer.c - 25/07/2000
 *
 * Gestion des timers
*/

#include <helium.h>

static He_vlist he_timer_list;
static int he_timer_uniq = 1;

typedef struct {
    int handle;
    He_timeout_proc proc;
    void *data;
    Ulong interval;     /* en millisecondes */
    double expiration;  /* timeofday absolu */
} He_timer_struct;


/*------------------- I N T E R F A C E - P U B L I Q U E --------------------*/

/*
 * Renvoie l'heure absolue sur un double en secondes,microsecondes
*/

double HeGetTime ()
{
  struct timeval tp;

  if (gettimeofday (&tp, (struct timezone *) NULL) == 0)
       return (double) tp.tv_sec + (double) tp.tv_usec / (double) 1E6;
  else return (double) 0;
}


/*
 * Ajout d'un timeout :
 *   la callback proc sera appele toutes les interval millisecondes.
 * HeAddTimeout renvoie un handle > 0 en cas de succes, -1 en cas d'erreur.
 * Le prototype de proc doit tre :
 *   int proc (int handle, void *data)
 * proc reoit le handle et data ; si proc renvoie FALSE, le timeout est
 * supprim ; si proc renvoie TRUE, le timeout est reconduit.
*/

int HeAddTimeout (Ulong interval, He_timeout_proc proc, void *data)
{
    He_timer_struct tmp;
    
    /* RQ interval >= 0 car type Ulong */
    
    if (proc == NULL) {
	HeError ("HeAddTimeout: proc is NULL\n");
	return -1;
    }
    
    tmp.handle = he_timer_uniq++;
    tmp.proc = proc;
    tmp.data = data;
    tmp.interval = interval;
    tmp.expiration = HeGetTime() + (double) interval / (double) 1E3;
    
    return (HeVListInsert (&he_timer_list, &tmp) < 0) ? -1 : tmp.handle;
}


/*
 * Supprime un timeout  partir de son handle.
 * Renvoie 0 succs, -1 erreur (mauvais handle).
*/

int HeRemoveTimeout (int handle)
{
    int i;
    
    if (handle <= 0) return -1;

    /* RQ la vlist est trie sur expiration, pas sur le handle : 
       on ne peut donc pas utiliser HeVListSearch */    
    for (i = 0; i < he_timer_list.nb; i++)
        if (((He_timer_struct*)he_timer_list.v)[i].handle == handle) 
            return HeVListRemove (&he_timer_list, i);
            
    return -1;
}


/*-------------------- F O N C T I O N S - P R I V E E S ---------------------*/

/*
 * Initialise le module.
*/

void HeTimerInit ()
{
    /* Liste de He_timer_struct trie dans l'ordre croissant du temps */
    HeVListInit (&he_timer_list, TRUE, 8, sizeof(He_timer_struct), 
	HeTimerCompar);
}


/*
 * Comparaison du temps pour la vlist
*/

int HeTimerCompar (void *v1, void *v2)
{
    return ((He_timer_struct*)v1)->expiration == 
           ((He_timer_struct*)v2)->expiration ?  0 :
           ((He_timer_struct*)v1)->expiration <  
           ((He_timer_struct*)v2)->expiration ? -1 : 1;
}


/*
 * Recherche de l'index pour insrer une expiration dans la vlist
 * Renvoie index >= 0.
*/

int HeTimerFind (double expiration)
{
    He_timer_struct elt;
    int index;

    elt.expiration = expiration;
    HeVListSearch (&he_timer_list, &elt, &index);
    return index;
}


/*
 * Renvoie le dlai entre l'appel et le prochain timeout.
 * renvoie 0 timeout, -1 pas de timeout.
*/

int HeGetNextTimeout (struct timeval *tp)
{
    double t = 0;
    
    if (he_timer_list.nb == 0) return -1;
    
    t = ((He_timer_struct*)he_timer_list.v)[0].expiration - HeGetTime();
    if (t < 0) t = 0;
    tp->tv_sec  = (int) t;
    tp->tv_usec = (int) ((t - tp->tv_sec) * 1E6);
    return 0;
}


/*
 * Excute le premier timeout de la liste s'il est coul.
 * Appel par HeSchedLoop().
*/

void HeTimerProcess ()
{
    He_timer_struct *v = he_timer_list.v, tmp;
    int res;
    
    if (he_timer_list.nb <= 0) return;
    if (v[0].expiration > HeGetTime()) return;  /* temps pas coul */
    
    /* On fait une copie pour viter tout effet de bord pendant l'appel de
       la callback */
    memmove (&tmp, &v[0], sizeof(He_timer_struct));
    
    /* Suppression du timeout 0 */
    HeVListRemove (&he_timer_list, 0);
       
    /* Appel callback */    
    res = (tmp.proc)(tmp.handle, tmp.data);
    
    /* Reconduction du timeout avec une nouvelle expiration. */
    if (res) {  
        tmp.expiration = HeGetTime() + (double) tmp.interval / (double) 1E3;
	HeVListInsert (&he_timer_list, &tmp);
    }
}


