/*
 * The Helium Toolkit
 *
 * Copyright (C) 1996-2000 Edouard Thiel <Edouard.Thiel@lif.univ-mrs.fr>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License; see http://www.lif-sud.univ-mrs.fr/~thiel/helium/LICENSE
*/

/*
 * gui/vlist.c - 29/03/2000
 *
 * Gestion de listes sans trous, tries ou non.
 *
 * La gestion mmoire est automatique (realloc avec un grain).
 *   - mode tri : tous les elt sont diffrents ; recherche dychotomique
 *   - mode non tri : les elts peuvent se rpter ; recherche itrative.
 *
 * On utilise la fonction memmove(dest,src,nbyte) qui recopie des zones
 * de mmoire pouvant se chevaucher.
 *
 * On peut stocker des entiers, rels, struct, pointeurs, etc. Il faut
 * faire attention dans le cas de struct aux problmes d'alignement
 * mmoire ; l'idal est que le struct ait une taille multiple de 4.
*/

#include <helium.h>


/*------------------- I N T E R F A C E - P U B L I Q U E --------------------*/

/*
 * EXEMPLES : soit le struct my_struct dfini par :
 * --------
 *            typedef struct { int cle, val1, val2; } my_struct;
 *
 *
 * 1) Utilisation de vlist pour stocker des struct my_struct
 *
 *    He_vlist my_list;
 *    HeVListInit (&my_list, TRUE, 256, sizeof(my_struct), my_compar);
 *
 *    int mycompar (void *v1, void *v2)
 *    {
 *        return ((my_struct*)v1)->cle == ((my_struct*)v2)->cle ?  0 :
 *               ((my_struct*)v1)->cle <  ((my_struct*)v2)->cle ? -1 : 1;
 *    }
 *
 *    my_struct *my_find (int cle)
 *    {
 *        my_struct elt;
 *        int index, res;
 *
 *        elt.cle = cle;
 *        res = HeVListSearch (&my_list, &elt, &index);
 *        return (res == 0) ? &((my_struct*)my_list.v)[index] : NULL;
 *    }
 *
 *
 * 2) Utilisation de vlist pour stocker des adresses my_struct*
 *
 *    He_vlist my_list;
 *    HeVListInit (&my_list, TRUE, 256, sizeof(my_struct*), my_compar);
 *
 *    int mycompar (void *v1, void *v2)
 *    {
 *        return (*(my_struct**)v1)->cle == (*(my_struct**)v2)->cle ?  0 :
 *               (*(my_struct**)v1)->cle <  (*(my_struct**)v2)->cle ? -1 : 1;
 *    }
 *
 *    my_struct *my_find (int cle)
 *    {
 *        my_struct st_elt, *elt = &st_elt;
 *        int index, res;
 *
 *        elt->cle = cle;
 *        res = HeVListSearch (&my_list, &elt, &index);
 *        return (res == 0) ? ((my_struct**)my_list.v)[index] : NULL;
 *    }
*/


/*
 * Initialise une structure He_vlist :
 *   - selon sorted, la liste est trie ou non.
 *   - grain est le pas de rallocation (>= 8).
 *   - size est la taille d'un elt en octets.
 *   - compar est la fonction de comparaison, renvoyant -1, 0 ou 1 : elle a
 *     les mme caractristiques que dans qsort() de stdlib.h. compar peut 
 *     tre  NULL en mode non tri si on n'appelle jamais HeVListSearch().
*/

void HeVListInit (He_vlist *list, int sorted, int grain, int size,
                  He_compar_proc compar)
{
    list->v = NULL;
    list->size = size;
    list->max = 0;
    list->nb = 0;
    if (grain < 8) grain = 8;
    list->grain = grain;
    list->inf = 4;
    list->sup = grain*2;
    list->sorted = sorted;
    list->compar = compar;
}


/*
 * Recherche d'un lment dans la liste.
 *   - p_elt est l'adresse d'un pointeur sur un struct dans lequel 
 *     on a mis le champ correspondant  la valeur recherche.
 *   - *ret_index est l'index trouv, sinon l'index o inserer.
 *     (sert en particulier pour HeVListInsert()).
 * Renvoie 0 trouv, -1 non trouv.
*/

int HeVListSearch (He_vlist *list, void *p_elt, int *ret_index)
{
    *ret_index = 0;
    if (list->compar == NULL || list->nb == 0) return -1;
    
    if (list->sorted) {  /* Recherche dychotomique */
    
        int m1 = 0,                 /* borne inf */
            m2 = list->nb,          /* borne sup PAR EXCES */
            mid = 0,                /* milieu de l'intervalle */
            res = 0;          

        while (m2-m1 > 0) {

            mid = (m1+m2) / 2;    /* Milieu tq m1 <= mid < m2 */
            res = (list->compar)(p_elt, (char*)list->v + mid * list->size);

            if      (res < 0) m2 = mid;
            else if (res > 0) m1 = mid+1;
            else { *ret_index = mid; return 0; } /* bingo */
        }
        *ret_index = (res > 0) ? mid+1 : mid;    /* FIXED 04/03/2007 */
        return -1;
    
    } else {  /* Recherche itrative */
    
        int i;
        for (i = 0; i < list->nb; i++)
            if ((list->compar)(p_elt, (char*)list->v + i * list->size) == 0) {
                *ret_index = i;
                return 0;
            }
       *ret_index = list->nb;
       return -1;
    }
}


/*
 * Insre un lment selon le mode (tri ou non) par recopie.
 * En mode tri, vrifie l'unicit.
 * Se charge d'agrandir la liste au besoin.
 * Renvoie index >=0 si succs, -1 erreur.
*/

int HeVListInsert (He_vlist *list, void *p_elt)
{
    if (HeVListLonger (list) < 0) return -1;
    
    if (list->sorted) {  /* Insertion avec dcalage */
    
        int index;
        if (HeVListSearch (list, p_elt, &index) == 0) return -1;
        
        /* Dcalage [index..nb[ de 1 vers la droite */
        if (index < list->nb)
        memmove ( (char*)list->v + list->size * (index+1),
                  (char*)list->v + list->size * index,
                                   list->size * (list->nb - index) );
        
        /* Recopie de p_elt  la case [index] */
        memmove ( (char*)list->v + list->size * index,
                  p_elt, list->size );

        list->nb++;
        return index;
        
    } else { /* On insre  la fin sans aucun test */
    
        int index = list->nb++;

        /* Recopie de p_elt  la case [index] */
        memmove ( (char*)list->v + list->size * index,
                  p_elt, list->size );

        return index;
    }
}


/*
 * Supprime un lment selon le mode (tri ou non).
 * Se charge de rduire la liste au besoin.
 * Renvoie 0 succs, -1 erreur.
*/

int HeVListRemove (He_vlist *list, int index)
{
    if (index < 0 || index >= list->nb) return -1;
    
    if (list->sorted) {  /* Suppression avec dcalage */

        /* Dcalage [index+1..nb[ de 1 vers la gauche */
        if (index < list->nb-1)
        memmove ( (char*)list->v + list->size * index,
                  (char*)list->v + list->size * (index+1),
                                   list->size * (list->nb-1 - index) );
                    
	list->nb--;
                
    } else {  /* Suppression sans trou */

        /* Recopie du dernier elt  la case [index] */
        if (index < list->nb-1)
        memmove ( (char*)list->v + list->size * index,
                  (char*)list->v + list->size * (list->nb-1), 
                  list->size );
                  
	list->nb--;
    }
    
    if (HeVListShorter (list) < 0) return -1;
    return 0;
}


/*
 * RAZ liste : met nb  0 et raccourcit max  grain
 * renvoie 0 succs, <0 erreur
*/

int HeVListRAZ (He_vlist *list)
{
    void **tmp;
    int allocsize;

    /* Zero lments */
    list->nb = 0;
    
    if (list->max > list->grain) {
    
        /* ici list->v != NULL donc on peut appeler realloc */
        allocsize = list->grain * list->size;
        tmp = realloc (list->v, allocsize);
        
        if (tmp == NULL) return -1;
        
        list->v = tmp;
	list->max = list->grain;
    }
    return 0;
}

/*-------------------- F O N C T I O N S - P R I V E E S ---------------------*/

/*
 * Rallonge la liste si besoin
 * renvoie 0 succs, <0 erreur
*/

int HeVListLonger (He_vlist *list)
{
    void **tmp;
    int allocsize;

    /* Test si rallonge inutile */
    if (list->max - list->nb > list->inf) return 0;

    /* Certains compilos ne supportent pas de realloc (NULL, ..); */
    allocsize = (list->max + list->grain) * list->size;
    if (list->v == NULL)
         tmp = malloc (allocsize);
    else tmp = realloc (list->v, allocsize);

    /* Test succes */
    if (tmp == NULL) {
	HeError ("HeVListLonger: realloc error for %d bytes\n", allocsize);
	return -ENOMEM;
    }

    /* Succes, m-a-j */
    list->v = tmp;
    list->max += list->grain;
    return 0;
}


/*
 * Raccourci la liste si besoin
 * renvoie 0 succs, <0 erreur
*/

int HeVListShorter (He_vlist *list)
{
    void **tmp;
    int allocsize;

    /* Test si raccourci inutile */
    if (list->max - list->nb < list->sup || list->max <= list->grain) return 0;

    allocsize = (list->max - list->grain) * list->size;
    tmp = realloc (list->v, allocsize);

    /* Test succes */
    if (tmp == NULL) {
	HeError ("HeVListShorter: realloc error for %d bytes\n", allocsize);
	return -ENOMEM;
    }

    /* Succes, m-a-j */
    list->v = tmp;
    list->max -= list->grain;
    return 0;
}


